/*
 * Copyright (c) 2024 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <string.h>
#include <stdint.h>
#include <hal/nrf_ccm.h>
#include <zephyr/kernel.h>
#include <zephyr/ztest.h>
#include "common.h"

#define CCM NRF_CCM00
#define ENABLE_DATA_PRINTOUT false

enum protocol_t { BLE_MODE = CCM_MODE_PROTOCOL_Ble, IEEE_MODE = CCM_MODE_PROTOCOL_Ieee802154 };

static void run_ccm(uint32_t key[], uint32_t nonce[], job_t *p_injob, job_t *p_outjob,
    unsigned maclen, uint8_t adatamask, bool expect_error, bool expect_macerror,
    bool decrypt, enum protocol_t mode_protocol)
{
  nrf_ccm_enable(CCM);

  nrf_ccm_event_clear(CCM, NRF_CCM_EVENT_END);
  nrf_ccm_event_clear(CCM, NRF_CCM_EVENT_ERROR);

  nrf_ccm_adatamask_set(CCM, adatamask);

  nrf_ccm_key_set(CCM, key);
  nrf_ccm_nonce_set(CCM, nonce);

  nrf_ccm_in_ptr_set(CCM, (nrf_vdma_job_t const *)p_injob);
  nrf_ccm_out_ptr_set(CCM, (nrf_vdma_job_t const *)p_outjob);

  if (ENABLE_DATA_PRINTOUT) {
    printk("Key in registers:\n");
    print_uint32_array(key, 4);
    printk("Nonce in registers:\n");
    print_uint32_array(nonce, 4);
    printk("Input job:\n");
    print_job(p_injob, true);
  }

  unsigned mac_config_val = maclen ? ((maclen - 2) / 2) << CCM_MODE_MACLEN_Pos : 0;

  CCM->MODE = mac_config_val | (mode_protocol << CCM_MODE_PROTOCOL_Pos);
  if (decrypt) {
    CCM->MODE |= (CCM_MODE_MODE_Decryption << CCM_MODE_MODE_Pos);
  } else {
    CCM->MODE |= (CCM_MODE_MODE_Encryption << CCM_MODE_MODE_Pos);
  }

  nrf_ccm_task_trigger(CCM, NRF_CCM_TASK_START);

  while (!nrf_ccm_event_check(CCM, NRF_CCM_EVENT_END) && !nrf_ccm_event_check(CCM, NRF_CCM_EVENT_ERROR))
  {
    k_busy_wait(1);
  }

  if (ENABLE_DATA_PRINTOUT) {
    printk("Processing done\n");
    printk("Output job:\n");
    print_job(p_outjob, true);
  }

  zassert_equal(CCM->EVENTS_ERROR, (uint32_t)expect_error, "CCM EVENTS_ERROR is %u but expected %u!",
      CCM->EVENTS_ERROR, expect_error);
  zassert_true(!decrypt || CCM->MACSTATUS == (uint32_t)!expect_macerror,
      "MACSTATUS %u is wrong, decrypt is %u expect_macerror is %u!\n", CCM->MACSTATUS, decrypt,
      expect_macerror);

  nrf_ccm_event_clear(CCM, NRF_CCM_EVENT_END);
  nrf_ccm_event_clear(CCM, NRF_CCM_EVENT_ERROR);

  nrf_ccm_disable(CCM);
}

static void print_and_verify_bytearray(uint8_t *p_vals, uint8_t *p_expected_vals, unsigned size)
{
  if (ENABLE_DATA_PRINTOUT) {
    printk("0x");
  }
  for (size_t i = 0; i < size; i++) {
    if (ENABLE_DATA_PRINTOUT) {
      printk("%02X", p_vals[i]);
    }
    zassert_equal(p_vals[i], p_expected_vals[i], "Byte %d (0x%02X) does not match the expected output (0x%02X)!",
        i, p_vals[i], p_expected_vals[i]);
  }
  printk(" matches the expected output!\n");
}

static void build_ccm_test_case_and_run(uint8_t K[], uint8_t N[], uint8_t *P, uint16_t P_size, uint16_t mlen,
    uint8_t *A, uint16_t alen, uint8_t *C, uint16_t out_data_len, unsigned mac_len,
    uint8_t adata_mask, bool expect_error, bool expect_macerror, bool decrypt,
    enum protocol_t mode_protocol, uint16_t *p_mlen_out, bool expect_out_data_len_mismatch)
{
  uint8_t key_reverse[16];
  uint8_t nonce_reverse[16];
  reverse_bytearray(K, key_reverse, 16);
  reverse_bytearray(N, nonce_reverse, 16);
  uint32_t *nonce = (uint32_t *)nonce_reverse;
  uint32_t *key = (uint32_t *)key_reverse;

  job_t ccm_injob[] = {
      {(uint8_t *)&alen, sizeof(alen), CCM_DMA_ATTR_ALEN},
      {(uint8_t *)&mlen, sizeof(mlen), CCM_DMA_ATTR_MLEN},
      {(uint8_t *)A, alen, CCM_DMA_ATTR_ADATA},
      {(uint8_t *)P, P_size, CCM_DMA_ATTR_MDATA},
      {0x0, 0x0} // Marking end of job list
  };

  uint16_t alen_out = 0xffff;
  uint16_t mlen_out = 0xffff;
  uint8_t adata_out[alen];
  memset(adata_out, 0xff, alen);
  const uint32_t PAINT = 0xCABACABA;
  uint8_t out_data[out_data_len + 2 * sizeof(uint32_t)];
  memset(out_data, 0xff, out_data_len + 2 * sizeof(uint32_t));
  uint32_t *p_inside_paint = (uint32_t *)&out_data[0];
  uint32_t *p_outside_paint = (uint32_t *)&out_data[sizeof(out_data) - sizeof(uint32_t)];
  *p_inside_paint = PAINT;
  *p_outside_paint = PAINT;
  memset(&out_data[sizeof(uint32_t)], 0xff, out_data_len);

  job_t ccm_outjob[] = {
      {(uint8_t *)&alen_out, sizeof(alen_out), CCM_DMA_ATTR_ALEN},
      {(uint8_t *)&mlen_out, sizeof(mlen_out), CCM_DMA_ATTR_MLEN},
      {(uint8_t *)adata_out, sizeof(adata_out), CCM_DMA_ATTR_ADATA},
      {(uint8_t *)&out_data[sizeof(uint32_t)], out_data_len, CCM_DMA_ATTR_MDATA},
      {0x0, 0x0} // Marking end of job list
  };

  run_ccm(key, nonce, ccm_injob, ccm_outjob, mac_len, adata_mask, expect_error, expect_macerror, decrypt,
      mode_protocol);
  zassert_equal(alen_out, sizeof(adata_out), "Header data length (0x%02X) does not match expected size (0x%02X)\n",
      adata_out, sizeof(adata_out));
  zassert_equal(*p_inside_paint, PAINT, "Bufer overflow\n");
  zassert_equal(*p_outside_paint, PAINT, "Bufer overflow\n");
  if (!expect_macerror && !expect_error) {
    zassert_true(expect_out_data_len_mismatch || (mlen_out == out_data_len),
        "Out data length (0x%02X) does not match expected size (0x%02X)\n", mlen_out, out_data_len);
    if (decrypt) {
      printk("Decrypted data: ");
    } else {
      printk("Encrypted data: ");
    }
    print_and_verify_bytearray(&out_data[sizeof(uint32_t)], C, out_data_len);
    printk("Header (Adata): ");
    print_and_verify_bytearray(adata_out, A, alen_out);
  }

  if (p_mlen_out != NULL) {
    *p_mlen_out = mlen_out;
  }
}

static const bool DECRYPT = true;
static const bool ENCRYPT = false;
static const bool EXPECT_MAC_ERROR = true;
static const bool EXPECT_ERROR = true;

ZTEST(nrf_ccm_tests, test_ccm_12_encr_ble_255b_w_adata)
{
	printk("Encrypt 255 byte payload with adata in BLE mode\n"); // hardware will return success but not encrypt
	// anything and return a 0 len payload

	uint8_t K[16] = {0x99, 0xAD, 0x1B, 0x52, 0x26, 0xA3, 0x7E, 0x3E, 0x05, 0x8E, 0x3B, 0x8E, 0x27, 0xC2, 0xC6, 0x66};
	uint8_t N[16] = {0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x80, 0x24, 0xAB, 0xDC, 0xBA, 0xBE, 0xBA, 0xAF, 0xDE};
	uint8_t A[] = {0x0E};
	uint8_t P[] = {
    // 0xaa,  0x1,  0x2,  0x3,  0x4,  0x5,  0x6,  0x7,  0x8,  0x9,  0xa,  0xb,  0xc,  0xd,  0xe,  0xf,  0x10,
			0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21,
			0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32,
			0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43,
			0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52, 0x53, 0x54,
			0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65,
			0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76,
			0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
			0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
			0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
			0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba,
			0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb,
			0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc,
			0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed,
			0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa
  };
	uint8_t C[] = {
 0x7C,  0x62,  0xA6,  0x14,  0x65,  0x52,  0x1D,  0x82,  0x1B,  0x67,  0xF3,  0xEA,  0x76,  0x74,  0x75,  0x09,  0x3B,
 0xC5,  0x47,  0x58,  0x68,  0x78,  0x4A,  0x83,  0xF7,  0xEC,  0x29,  0x6C,  0x61,  0x1E,  0x0C,  0x71,  0xC8,  0xBC,
 0xE7,  0xAF,  0xEF,  0x95,  0xD4,  0x3C,  0x2B,  0xBB,  0x24,  0xF3,  0xF7,  0xB3,  0xA7,  0xA6,  0x39,  0x8C,  0x8E,
 0x82,  0xD1,  0xB3,  0x68,  0x4D,  0x00,  0x13,  0x10,  0x7B,  0x6A,  0x96,  0x32,  0xA1,  0xA4,  0xAF,  0x8F,  0x2D,
 0x2C,  0x96,  0xA6,  0xB5,  0x74,  0x77,  0x3F,  0x2F,  0x93,  0x34,  0x2C,  0x8B,  0x22,  0xCD,  0xD4,  0xBE,  0x2E,
 0x1C,  0x3F,  0xA4,  0x33,  0x38,  0x21,  0xFA,  0x99,  0xFE,  0x8B,  0x89,  0xC1,  0x53,  0x55,  0xC7,  0x64,  0x24,
 0xAA,  0x19,  0x07,  0x47,  0x15,  0x68,  0x9B,  0x1F,  0x3F,  0x91,  0x49,  0xF5,  0xF7,  0x94,  0xCE,  0xB9,  0xF9,
 0x00,  0xD0,  0xDB,  0x9D,  0x78,  0x94,  0x45,  0xD8,  0x0A,  0xED,  0x4B,  0x38,  0x6A,  0x44,  0xED,  0xB9,  0x3B,
 0xCF,  0xF6,  0xA0,  0x9B,  0x8A,  0xBC,  0xC2,  0x57,  0xFF,  0xB4,  0xA3,  0x30,  0x7E,  0x73,  0x44,  0xD9,  0x96,
 0x52,  0x38,  0x44,  0x8F,  0x39,  0x18,  0x2A,  0x2E,  0xE4,  0xAC,  0xCF,  0x8D,  0xD4,  0xCB,  0x54,  0xBD,  0xFA,
 0xD3,  0xF3,  0xC6,  0x5F,  0x3D,  0x49,  0xFD,  0x73,  0x95,  0x91,  0xCF,  0x76,  0x7F,  0xBD,  0xAB,  0x6D,  0xD6,
 0xD9,  0xA6,  0x46,  0xD9,  0x49,  0xDC,  0xEB,  0x82,  0x84,  0x37,  0x1F,  0xC4,  0x58,  0x8A,  0x6F,  0x24,  0x75,
 0x6C,  0xE8,  0xDE,  0x82,  0x49,  0xA2,  0x97,  0x3B,  0xF9,  0x6F,  0x8E,  0x3A,  0x8E,  0x4D,  0xD5,  0x3B,  0x7B,
 0xAD,  0xF8,  0xFD,  0x8E,  0x66,  0x6B,  0x4F,  0x95,  0x88,  0xC5,  0x85,  0x3A,  0x42,  0x92,  0xFB,  0x40,  0xD6

  };
  printk("P size: %d, C size: %d\n", sizeof(P), sizeof(C));
	unsigned mac_len = 4;
	uint8_t adata_mask = 0xE3;
	build_ccm_test_case_and_run(K, N, P, sizeof(P), sizeof(P), A, sizeof(A), C, sizeof(C), mac_len, adata_mask, false,
			false, ENCRYPT, BLE_MODE, NULL, false);
}

ZTEST_SUITE(nrf_ccm_tests, NULL, NULL, NULL, NULL, NULL);
